/*
 *
 *iXQ - cocoa-based icq client for MacOS X
 *Copyright (C) 2003  Franois-Marie Grattepain, Philippe Champeaux
 *
 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.
 *
 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.
 *
 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
//
//  SMServiceMgr.m
//  iXQnew
//
//  Created by Franois-Marie Grattepain on Wed Jan 23 2002.
//  Copyright (c) 2002 iXQ-Team. All rights reserved.
//

#import "SMServiceMgr.h"
#import "SMDemultiplexer.h"
#import "SMConnectionPackets.h"
#import "SMDataPackets.h"
#import "SMErrorPackets.h"
#import "SMConnection.h"

@implementation SMServiceMgr

// (de)initialization

// new kernel instantiation: given two ports to communicate with the GUI, and the user's UIN
+ (void) managerWithData:(NSArray*)ports
{
    NSConnection* guiConnection;
    SMServiceMgr* manager;
    NSAutoreleasePool* releasePool;

    // the release pool prevent us from memory leaks :)
    releasePool = [[NSAutoreleasePool alloc] init];
    // NSLog(@"SMServiceMgr init");

    // opening the connection to the GUI
    guiConnection = [NSConnection connectionWithReceivePort:[ports objectAtIndex:0] sendPort:[ports objectAtIndex:1]];
    manager = (SMServiceMgr*)[[self alloc] initWithParent:(id)[guiConnection rootProxy] andUIN:[ports objectAtIndex:2]];
    [(id)[guiConnection rootProxy] setServer:manager];

    [manager connect];
    [manager release];
    
    // to be sure the thread won't be existed at the end of the method
    [[NSRunLoop currentRunLoop] run];

    [releasePool release];
    
    return;
}

- (void) dealloc
{
    [connectionPkHandler release];
    [dataPkHandler release];
    [errorPkHandler release];
    [demultiplexer release];
    [guiConnectionRoot release];
}

// initialization of a new instance with its connected GUI Interface
-(id)initWithParent:(id)parent andUIN:(NSString*)userUIN
{
    self = [super init];
    if (self) {

        // initialization of each linked class
        // then the classes register to the demultiplexer
        // so that it can directly access them 
        demultiplexer = (SMDemultiplexer*)[[SMDemultiplexer alloc] init];

        connectionPkHandler = (SMConnectionPackets*)[[SMConnectionPackets alloc] initWithDemultiplexer:demultiplexer andManager:self];
        [demultiplexer registerConnectionPkHandler:connectionPkHandler];

        dataPkHandler = (SMDataPackets*)[[SMDataPackets alloc] initWithDemultiplexer:demultiplexer andManager:self];
        [demultiplexer registerDataPkHandler:dataPkHandler];

        errorPkHandler = (SMErrorPackets*)[[SMErrorPackets alloc] initWithDemultiplexer:demultiplexer andManager:self];
        [demultiplexer registerErrorPkHandler:errorPkHandler];

        guiConnectionRoot = (id <connectionProtocol>)[parent retain];

        uin = [userUIN intValue];
        connected = NO;
        seqNum = time(0);
    }
    return self;
}

// accessors

- (unsigned int) seqNum
{
    return (++seqNum);
}

- (unsigned int) newStatus
{
    return (requestedStatus - 1);
}

// gui-data

// called from the kernel to obtain the contact list
- (NSSet*) contacts
{
    NSSet* contacts = [guiConnectionRoot contacts];
    return contacts;
}

// called from the kernel to obtain the visible list
// needed at statuschange/connection to invisible status
- (NSSet*) visible
{	
    NSSet *set = [guiConnectionRoot visibleList];
    return set;
}

// called from the kernel to obtain the invisible list
// needed at statuschange/connection to a visible status
- (NSSet*) invisible
{
    NSSet *set = [guiConnectionRoot invisibleList];
    return set;
}

// gui-callbacks (outgoing)

- (void) connectionError:(NSString*)error
{
    NSDictionary *dict=[NSDictionary dictionaryWithObjectsAndKeys: [NSString stringWithFormat:@"%d", uin], @"uin", error, @"message", nil];
    
    [guiConnectionRoot sendNotification:@"loginError" object:self userInfos:dict];
    if (connected)
        [self setStatus:0];
}

// passes a user found from the kernel to the gui
- (void) userFound:(SMUserInfos*)ui
{
    NSDictionary *dict = [NSDictionary dictionaryWithObjectsAndKeys: [NSString stringWithFormat:@"%d", uin], @"uin", [ui toDict], @"SMUserInfos", nil];
			
    [guiConnectionRoot sendNotification:@"userFound" object:self userInfos:dict];
}

// passes the full info for a contact from the kernel to the gui
- (void) gotFullUserInfo:(SMUserInfos*)ui 
{
    NSDictionary *dict=[NSDictionary dictionaryWithObjectsAndKeys: [NSString stringWithFormat:@"%d", uin], @"uin", ui, @"SMUserInfos", nil];
                        
    [guiConnectionRoot sendNotification:@"userDetailsForUIN" object:self userInfos:dict];
}

// informs the gui when a contact changes his status
- (void) userWithUIN:(int)UIN changedStatusTo:(int)newStatus withIP:(NSString*)ipAddress withCap:(NSString*)capabilities withClient:(char)vers localIP:(NSString*)inLocalIP port:(int)inPort cookie:(int)inCookie
{
    NSDictionary *dict=[NSDictionary dictionaryWithObjectsAndKeys:
        [NSString stringWithFormat:@"%d", uin], @"uin",
        [NSString stringWithFormat:@"%d", UIN], @"contactUIN",
        ipAddress, @"ipAddress",    
        capabilities, @"capabilities",    
	  [NSNumber numberWithInt:vers], @"clientversion",
	  [NSNumber numberWithInt:inCookie], @"cookie",
	  [NSNumber numberWithInt:inPort], @"port",
	  inLocalIP, @"localIP",
        [NSNumber numberWithInt:newStatus], @"contactStatus", nil];
  
//    NSLog([NSString stringWithFormat:@"%d changed status to : %d", UIN, newStatus]);
    [guiConnectionRoot sendNotification:@"contactChangedStatus" object:self userInfos:dict];
}

// passes a received message to the gui
- (void) receivedMessage:(NSData*)content fromUIN:(int)UIN
{
    SMMessage *newMessage=[[SMMessage alloc] initWithData:[content retain] srcUIN:[NSString stringWithFormat:@"%d", UIN] 
        destUIN:[NSString stringWithFormat:@"%d", uin]];
    NSDictionary *dict=[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithFormat:@"%d", uin], @"uin", newMessage, @"message", nil];
    
    //NSLog([NSString stringWithFormat:@"Received Message From : %d : \n%@", UIN, content]);
    [guiConnectionRoot sendNotification:@"newMessageReceived" object:self userInfos:dict];
}

// passes a received message to the gui
- (void) receivedServerSideList:(NSArray*)inUins
{
  NSDictionary *dict = [NSDictionary dictionaryWithObjectsAndKeys: [NSString stringWithFormat:@"%d", uin], @"uin", inUins, @"UINList", nil];

  [guiConnectionRoot sendNotification:@"serverSideListReceived" object:self userInfos:dict];
}



// informs the gui of a successful status change
- (void) statusChanged:(BOOL)success
{	
    NSDictionary *dict;
    if (success&&!connected)
    {
        dict=[NSDictionary dictionaryWithObjectsAndKeys: [NSString stringWithFormat:@"%d", uin], @"uin", nil];
        [guiConnectionRoot sendNotification:@"connectionCompleted" object:self userInfos:dict];

        connected=(requestedStatus!=0);
        connectionAlive = YES;
        kaTimer = [NSTimer scheduledTimerWithTimeInterval:10 target:self selector:@selector(testConnection:) userInfo:nil repeats:YES];
    }
    
    dict=[NSDictionary dictionaryWithObjectsAndKeys: [NSString stringWithFormat:@"%d", uin], @"uin", 
        [NSNumber numberWithInt:requestedStatus], @"status", nil];
    
    [guiConnectionRoot sendNotification:@"statusChanged" object:self userInfos:dict];
}

- (void) authorizationRequestFrom:(int)UIN message:(NSString*)msg
{
    NSDictionary *dict=[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithFormat:@"%d", uin], @"uin", [NSString stringWithFormat:@"%d", UIN], @"requesterUIN",  msg, @"message",nil];

    [guiConnectionRoot sendNotification:@"authorisationRequest" object:self userInfos:dict];
}

- (void) receivedURL:(NSString*)URL description:(NSString*)desc fromUIN:(int)UIN
{
    NSString* message = [NSString stringWithFormat:@"URL:'%@'\nDescription:'%@'", URL, desc];
    SMMessage *newMessage=[[SMMessage alloc] initWithData:[message dataUsingEncoding:nil] srcUIN:[NSString stringWithFormat:@"%d", UIN] destUIN:[NSString stringWithFormat:@"%d", uin]];
    NSDictionary *dict=[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithFormat:@"%d", uin], @"uin", newMessage, @"message", nil];
    
    //NSLog([NSString stringWithFormat:@"Received Message From : %d : \n%@", UIN, content]);
    [guiConnectionRoot sendNotification:@"newMessageReceived" object:self userInfos:dict];
}

- (void) userAddedYou:(int)UIN
{
    NSString* message = [NSString stringWithFormat:@"This contact with UIN %d has added you to his/her contact list", UIN];
    SMMessage *newMessage=[[SMMessage alloc] initWithData:[message dataUsingEncoding:nil] srcUIN:[NSString stringWithFormat:@"%d", UIN] destUIN:[NSString stringWithFormat:@"%d", uin]];
    NSDictionary *dict=[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithFormat:@"%d", uin], @"uin", newMessage, @"message", nil];
    
    //NSLog([NSString stringWithFormat:@"Received Message From : %d : \n%@", UIN, content]);
    [guiConnectionRoot sendNotification:@"newMessageReceived" object:self userInfos:dict];

}

- (void) infoModificationSucceeded
{
    NSDictionary *dict=[NSDictionary dictionaryWithObjectsAndKeys: [NSString stringWithFormat:@"%d", uin], @"uin", nil];
    
    [guiConnectionRoot sendNotification:@"infoModificationSucceeded" object:self userInfos:dict];
}

// kernel-functions (incoming)

- (void) initiateConnection
{
    NSDictionary* connectionInfos;
    //get the password, the host and the port
    connectionInfos = [guiConnectionRoot connectionInfos];
    [connectionPkHandler connectUIN:uin password:[connectionInfos objectForKey:@"password"]];
}

- (void) connect
{
    NSDictionary* connectionInfos;

    //We should connect with the new status (SMConnectionPacket)
    //->SMDemultiplexer : Sockets Initialization
    //get the password, the host and the port
    connectionInfos = [guiConnectionRoot connectionInfos];

    // dont know why it was this way but its a bug when connecting invisible
    // requestedStatus = ([[connectionInfos objectForKey:@"loginstatus"] intValue]+1)%8-1;     
    requestedStatus = ([[connectionInfos objectForKey:@"loginstatus"] intValue]);   

    [demultiplexer connectSocketToHost:[connectionInfos objectForKey:@"host"] onPort:[[connectionInfos objectForKey:@"port"] intValue]];
}

- (void) queryOurInfo
{
    [dataPkHandler queryOurInfo:uin];
}

- (void) modifyOurInfo:(SMUserInfos*)ui
{
    [dataPkHandler modifyOurInfo:uin newInfo:ui];
}
// called from the gui when we want to change our current status
- (oneway void) setStatus:(unsigned int)newStatus
{
    // NSLog([NSString stringWithFormat:@"Receiving a status change order for status : %d", newStatus]);
    requestedStatus = newStatus;
    //If we don"t request something hard to do (go offline while not connected :^p)
    if (!connected && newStatus==0)
        [self statusChanged:NO];
    //If we don't switch offline/online
    else if (connected && newStatus > 0) //&&newStatus!= @"Offline"
        //We just change status (SMDataPackets) 
        [dataPkHandler changeStatus];
    else if (newStatus==0)
        //disconnect (close the sockets/clean the thread)
    {
      [kaTimer invalidate];
	  [demultiplexer closeSocket];
	  [self statusChanged:NO];
	  connected = NO;
	  [self release];
	  [NSThread exit];
    }
    else
    {
        //We should connect with the new status (SMConnectionPacket)
        //->SMDemultiplexer : Sockets Initialization
        NSDictionary* connectionInfos;
        //get the password, the host and the port
        //NSLog([NSString stringWithFormat:@"toto (%d/%d)", newStatus, requestedStatus]);
        connectionInfos = [guiConnectionRoot connectionInfos];
        [demultiplexer connectSocketToHost:[connectionInfos objectForKey:@"host"] onPort:[[connectionInfos objectForKey:@"port"] intValue]];
    }
}

// called from the gui to search for a user
- (void) searchContact:(SMUserInfos*)ui
{
    if ([[ui UIN] isEqualToString:@""])
    {
    // 	[dataPkHandler searchForUserWithNickname:[ui nickname] firstName:@"" lastname:@"" fromUIN:uin];
    //  we always use our extendet search implementation now
        [dataPkHandler searchExtendedForUser:ui fromUIN:uin];
    }
    else
        [dataPkHandler searchForUIN:[[ui UIN] intValue] withUIN:uin];
}

- (void) retrieveServerSideList
{
	[dataPkHandler requestServerSideContacts];
}


// called from the gui to request the full info for a contact
- (void) requestFullInfoForUIN:(NSString*)suin
{
    [dataPkHandler requestFullInfoForUIN:[suin intValue] withUIN:uin];
}

- (void) requestOfflineMessages
{
//  NSLog(@"Requesting offline messages");
    [dataPkHandler requestOfflineMessagesForUIN:uin];
}

- (void) addUserToContactList:(SMUserInfos*)ui
{
    [dataPkHandler sendContactList:[NSSet setWithObject:ui]];
}

- (void) sendOfflineMessage:(NSData*)message toUIN:(NSString*)UIN
{
    [dataPkHandler sendOfflineMessage:message toUIN:[UIN intValue] from:uin];
}

- (void) sendMessage:(NSData*)message toUIN:(NSString*)UIN
{
    [dataPkHandler sendDataMessage:message toUIN:[UIN intValue] from:uin];
}

- (void) authorizationReplyTo:(int)UIN isGiven:(BOOL)given;
{
    [dataPkHandler sendAuthorizationReplyTo:UIN from:uin isGiven:given];
}

- (void) endOfOfflineMessages
{
  [dataPkHandler ackOfflineMessagesForUIN:uin];
  [dataPkHandler requestServerSideContacts];
}


- (void) addToInvisibleList:(NSSet*)list
{
    [list retain];
    [dataPkHandler sendUINList:list subtype:7];
    [list release];    
}

- (void) addToVisibleList:(NSSet*)list
{
    [list retain];
    [dataPkHandler sendUINList:list subtype:5];
    [list release];  
}

- (void) removeFromInvisibleList:(NSSet*)list
{
    [list retain];
    [dataPkHandler sendUINList:list subtype:8];
    [list release];
}

- (void) removeFromVisibleList:(NSSet*)list
{
    [list retain];
    [dataPkHandler sendUINList:list subtype:6];
    [list release];
}

- (void) testConnection:(NSTimer*)unused
{

  if (connectionAlive)
  {
//    NSLog(@"### Connection OK ###");
    connectionAlive = NO;
//    [dataPkHandler testWithSearchForUIN:(uin+1) withUIN:uin];
	[dataPkHandler testWithKeepAlive];
  }
  else {
  
    NSDictionary *dict=[NSDictionary dictionaryWithObjectsAndKeys:
        [NSString stringWithFormat:@"%d", uin], @"uin", 
        nil];
    
    [guiConnectionRoot sendNotification:@"connectionLost" object:self userInfos:dict];
    NSLog(@"### Connection failed ###");
	[unused invalidate];
  }
}

- (void) error
{

  NSDictionary *dict=[NSDictionary dictionaryWithObjectsAndKeys:
	[NSString stringWithFormat:@"%d", uin], @"uin", 
	nil];

  [guiConnectionRoot sendNotification:@"connectionLost" object:self userInfos:dict];
  NSLog(@"### Connection error ###");
}

- (void) connectionAlive
{
  connectionAlive = YES;
}


@end
